from twisted.cred import portal, checkers, credentials
from twisted.conch import error, avatar, recvline, interfaces as conchinterfaces
from twisted.conch.ssh import factory, userauth, connection, keys, session, common
from twisted.conch.insults import insults
from twisted.application import service, internet
from twisted.internet import protocol
from twisted.internet import reactor
from twisted.python import log, randbytes
from twisted.conch.ssh import transport
from twisted.conch.ssh.common import MP, NS, getMP

from zope import interface
from zope.interface import implements
import os
import re
import sys
import hashlib


from PythonSSH.DiffieHellmanFake import DiffieHellmanServer
from Crypto import Util
from PythonSSH.SSHUtil import *
from PythonSSH.PacketUtil import *
import base64


class ServerCommandTransport(transport.SSHServerTransport):
    SSH_MSG_KEXDH_REPLY = transport.MSG_KEXDH_REPLY
    
    def GetOurVersionStr(self):
        return self.ourVersionString

    def GetOtherVersionStr(self):
        return self.otherVersionString
            
    def GetOurKexInitPayload(self):
        """The packet we sent as part of the key initialization"""
        return self.ourKexInitPayload

    def GetOtherKexInitPayload(self):
        """The packet we received as part of the key initialization"""
        return self.otherKexInitPayload
            
    def __init__(self):
        #transport.SSHServerTransport.__init__(self)
        
        self.supportedKeyExchanges = self.GetSupportedKeyExchanges()
        self.supportedCompressions = self.GetSupportedCompressions()
        self.supportedMACs = self.GetSupportedMACs()
        self.supportedCiphers = self.GetSupportedCiphers()
        #self.username = username
        #self.password = password  
        #self.command = command
        self.sharedSecret = None       
        self.exchangeHash = None
        
        self.outIV = None
        self.outKey = None
        self.inIV = None
        self.inKey = None
        self.outInteg = None
        self.inInteg = None
        
        self.inAes = None
        self.outAes = None
        #log.startLogging(sys.stdout)
        

    def _keySetup(self, sharedSecret, exchangeHash):
        """
        Set up the keys for the connection and sends MSG_NEWKEYS when
        finished,

        @param sharedSecret: a secret string agreed upon using a Diffie-
                             Hellman exchange, so it is only shared between
                             the server and the client.
        @type sharedSecret: C{str}
        @param exchangeHash: A hash of various data known by both sides.
        @type exchangeHash: C{str}
        """
        if not self.sessionID:
            self.sessionID = exchangeHash
        initIVCS = self._getKey('A', sharedSecret, exchangeHash)
        initIVSC = self._getKey('B', sharedSecret, exchangeHash)
        encKeyCS = self._getKey('C', sharedSecret, exchangeHash)
        encKeySC = self._getKey('D', sharedSecret, exchangeHash)
        integKeyCS = self._getKey('E', sharedSecret, exchangeHash)
        integKeySC = self._getKey('F', sharedSecret, exchangeHash)
        outs = [initIVSC, encKeySC, integKeySC]
        ins = [initIVCS, encKeyCS, integKeyCS]
        if self.isClient: # reverse for the client
            log.msg('REVERSE')
            outs, ins = ins, outs
        self.setKeys(outs[0], outs[1], ins[0], ins[1], outs[2], ins[2])
        self.sendPacket(transport.MSG_NEWKEYS, '')


    def setKeys(self, outIV, outKey, inIV, inKey, outInteg, inInteg):
        """
        Set up the ciphers and hashes using the given keys,

        @param outIV: the outgoing initialization vector
        @param outKey: the outgoing encryption key
        @param inIV: the incoming initialization vector
        @param inKey: the incoming encryption key
        @param outInteg: the outgoing integrity key
        @param inInteg: the incoming integrity key.
        """
        import base64
        self.outIV = outIV
        self.outKey = outKey
        self.inIV = inIV
        self.inKey = inKey
        self.outInteg = outInteg
        self.inInteg = inInteg        
        self.nextEncryptions.setKeys(outIV, outKey, inIV, inKey, outInteg, inInteg)
            
    def sendPacket(self, messageType, payload):
        """
        Sends a packet.  If it's been set up, compress the data, encrypt it,
        and authenticate it before sending.

        @param messageType: The type of the packet; generally one of the
                            MSG_* values.
        @type messageType: C{int}
        @param payload: The payload for the message.
        @type payload: C{str}
        """
        payload = chr(messageType) + payload
        if self.outgoingCompression:
            payload = (self.outgoingCompression.compress(payload)
                       + self.outgoingCompression.flush(2))
        bs = self.currentEncryptions.encBlockSize
        # 4 for the packet length and 1 for the padding length
        totalSize = 5 + len(payload)
        lenPad = bs - (totalSize % bs)
        if lenPad < 4:
            lenPad = lenPad + bs
        packet = (struct.pack('!LB',
                              totalSize + lenPad - 4, lenPad) +
                  payload + randbytes.secureRandom(lenPad))
                  
        #add bz
        from Crypto.Cipher import AES
        if self.currentEncryptions.outCipType == 'aes128-cbc':
            if self.outAes is None:
                self.outAes = PacketUtil.CreateAESObject(self.outKey, 16, self.outIV, self.currentEncryptions.encBlockSize)
            macKey = self.outInteg[:20] + '\x00' * (64 - 20)
            encPacket = PacketUtil.EncryptAndMACPacket(self.outgoingPacketSequence, packet, self.outAes, macKey)
        else:
                    encPacket = (
                        self.currentEncryptions.encrypt(packet) +
                        self.currentEncryptions.makeMAC(
                            self.outgoingPacketSequence, packet))
        #end add bz                  
                  

                   
        self.transport.write(encPacket)
        self.outgoingPacketSequence += 1
        
    def getPacket(self):
        """
        Try to return a decrypted, authenticated, and decompressed packet
        out of the buffer.  If there is not enough data, return None.

        @rtype: C{str}/C{None}
        """
        
        from Crypto.Cipher import AES
        if self.currentEncryptions.inCipType == 'aes128-cbc':
            if self.inAes is None:
                self.inAes = PacketUtil.CreateAESObject(self.inKey, 16, self.inIV, self.currentEncryptions.decBlockSize)        
        
        bs = self.currentEncryptions.decBlockSize
        ms = self.currentEncryptions.verifyDigestSize
        if len(self.buf) < bs: return # not enough data
        if not hasattr(self, 'first'):
            if self.currentEncryptions.inCipType == 'aes128-cbc':
                first = PacketUtil.DecryptPacket(self.buf[:bs], self.inAes)
                #first = self.currentEncryptions.decrypt(self.buf[:bs])
            else:
                first = self.currentEncryptions.decrypt(self.buf[:bs])
        else:
            first = self.first
            del self.first
        packetLen, paddingLen = struct.unpack('!LB', first[:5])
        if packetLen > 1048576: # 1024 ** 2
            self.sendDisconnect(DISCONNECT_PROTOCOL_ERROR,
                                'bad packet length %s' % packetLen)
            return
        if len(self.buf) < packetLen + 4 + ms:
            self.first = first
            return # not enough packet
        if(packetLen + 4) % bs != 0:
            self.sendDisconnect(
                DISCONNECT_PROTOCOL_ERROR,
                'bad packet mod (%i%%%i == %i)' % (packetLen + 4, bs,
                                                   (packetLen + 4) % bs))
            return
        encData, self.buf = self.buf[:4 + packetLen], self.buf[4 + packetLen:]
        if self.currentEncryptions.inCipType == 'aes128-cbc':
            #packet = first + self.currentEncryptions.decrypt(encData[bs:])
            packet = first + PacketUtil.DecryptPacket(encData[bs:], self.inAes)
        else:
            packet = first + self.currentEncryptions.decrypt(encData[bs:])
     
        
        if len(packet) != 4 + packetLen:
            self.sendDisconnect(DISCONNECT_PROTOCOL_ERROR,
                                'bad decryption')
            return
        if ms:
            macData, self.buf = self.buf[:ms], self.buf[ms:]
            macKey = self.inInteg[:20] + '\x00' * (64 - 20)
            if not PacketUtil.VerifyMAC(macData, self.incomingPacketSequence, packet, macKey):
                self.sendDisconnect(DISCONNECT_MAC_ERROR, 'bad MAC')
                return
        payload = packet[5:-paddingLen]
        if self.incomingCompression:
            try:
                payload = self.incomingCompression.decompress(payload)
            except: # bare except, because who knows what kind of errors
                    # decompression can raise
                log.err()
                self.sendDisconnect(DISCONNECT_COMPRESSION_ERROR,
                                    'compression error')
                return
        self.incomingPacketSequence += 1
        return payload
        
        
    def ssh_KEX_DH_GEX_REQUEST_OLD(self, packet):
        if self.kexAlg == 'diffie-hellman-group1-sha1':
            if self.ignoreNextPacket:
                self.ignoreNextPacket = 0
                return
            # this is really KEXDH_INIT
            clientDHpublicKey, foo = getMP(packet)

            self.KeyExchangeReceivedInit(clientDHpublicKey, self.factory.publicKeys[self.keyAlg].blob())
        else:
            transport.SSHServerTransport.ssh_KEX_DH_GEX_REQUEST_OLD(self, packet)
            
           
            
